# Helpers for using PyVO and astropy, as used by the PyVO talk(s) given
# by GAVO Heidelberg.
#
# This code is in the public domain.

import contextlib
import functools
import itertools
import re
import tempfile
import traceback
import warnings

# astropy's votable code is overzealous in complaining about things
# -- that's worthless for a consumer, so let's turn it off.
warnings.filterwarnings('ignore', module="astropy.io.votable.*")

try:
	from astropy.vo.samp import SAMPIntegratedClient, SAMPProxyError
except ImportError:
	from astropy.samp import SAMPIntegratedClient, SAMPProxyError

import numpy as np
import pyvo
import requests

import os


def show_exception(func):
	"""decorates func such that any exceptions coming out of it are
	shown in the terminal (and then re-raised).
	"""
	def _(*args, **kwargs):
		try:
			return func(*args, **kwargs)
		except Exception:
			traceback.print_exc()
			raise

	return functools.update_wrapper(_, func)


def send_table_to(conn, dest_client_id, astropy_table, name="data"):
	"""sends astropy_table via SAMP.

	dest_client_id can be obtained with find_client.  This does the
	exchange through the file system and thus assumes that the 
	target client shares it with us.
	"""
	with samp_accessible(astropy_table) as table_url:
		message = {
			"samp.mtype": "table.load.votable",
			"samp.params": {
				"url": table_url,
				"name": name,
			}}
		conn.call_and_wait(dest_client_id, message, "10")


def send_product_to(conn, dest_client_id, data_url, mtype, name="data"):
	"""sends SAMP messages to load data.

	This is a helper for send_spectrum_to and send_image_to, which work
	exactly analogous to each other, except that the mtypes are different.

	If dest_client_id, this is a broadcast (and we don't wait for any
	responses).  If dest_client_id is given, we wait for acknowledgement
	by the receiver.
	"""
	message = {
		"samp.mtype": mtype,
		"samp.params": {
			"url": data_url,
			"name": name,
		}}
	if dest_client_id is None:
		conn.notify_all(message)
	else:
		conn.call_and_wait(dest_client_id, message, "10")


def send_spectrum_to(conn, dest_client_id, spectrum_url, name="data"):
	"""asks a spectrum client to open a remote spectrum via SAMP.

	Pass dest_client_id=None to do a broadcast.
	"""
	send_product_to(conn, 
		dest_client_id, 
		spectrum_url, 
		"spectrum.load.ssa-generic",
		name)


def send_image_to(conn, dest_client_id, image_url, name="data"):
	"""asks an image client to open a remote image via SAMP.

	Pass dest_client_id=None to do a broadcast.

	It's likely that anything that's not a FITS file won't work here,
	so try to avoid that.
	"""
	send_product_to(conn, 
		dest_client_id, 
		image_url, 
		"image.load.fits",
		name)


def find_client(conn, samp_name):
	"""returns the SAMP id of the client with samp.name samp_name.

	This will raise a KeyError if the client is not on the hub.
	"""
	for client_id in conn.get_registered_clients():
		if conn.get_metadata(client_id).get("samp.name")==samp_name:
			return client_id
	raise KeyError(samp_name)


@contextlib.contextmanager
def samp_accessible(astropy_table):
	"""a context manager making astropy_table available under a (file)
	URL for the controlled section.

	This is useful with uploads.
	"""
	handle, f_name = tempfile.mkstemp(suffix=".xml")
	with os.fdopen(handle, "w") as f:
		astropy_table.write(output=f,
			format="votable")
	try:
		yield "file://"+f_name
	finally:
		os.unlink(f_name)


@contextlib.contextmanager
def SAMP_conn(
		client_name="pyvo client", 
		description="A generic PyVO client",
		**kwargs):
	"""a context manager to give the controlled block a SAMP connection.

	The program will disconnect as the controlled block is exited.
	"""
	client = SAMPIntegratedClient(
		name=client_name,
		description=description,
		**kwargs)
	client.connect()
	try:
		yield client
	finally:
		client.disconnect()


# TODO: use resultset.fieldname_with_ucd instead
def get_name_for_ucd(ucd, table):
	"""returns the name of a column having ucd in table.

	If there are multiple such columns, a random one is returned.  If there
	are none, a key error is raised.
	"""
	ucd = ucd.lower()

	for col in table.columns.values():
		if col.meta.get("ucd", "").lower()==ucd:
			return col.name
	raise KeyError(ucd)


SOME_SQL_RESERVED_WORDS = set(
	"area box centroid circle coordsys distance exp log point region"
	" avg case cross current date day desc distinct double exists"
	" found full global group hour key left level max min month"
	" precision prior public real right second set size sum"
	" time timestamp to true upper user value when where year".split())


def quote_if_necessary(identifier):
	"""returns a delimited version of identifier if it doesn't look it would
	pass for a SQL regular identifier.

	We actually allow dots since we don't want to parse table references
	with schema names.  If someone is devious enough to break this
	with a simple dot, they have waived their moaning rights.

	We only check for a few of the most tempting SQL keywords, though.

	This function shouldn't really be necessary here as TAP operators ought
	to give pre-quoted identifiers in tap_schema and friends.  They don't
	yet, though.
	"""
	if re.match('".*"$', identifier):
		# we are already quoted
		return identifier

	elif (re.match("[A-Za-z][A-Za-z0-9_.]*$", identifier) 
			and not identifier.lower() in SOME_SQL_RESERVED_WORDS):
		return identifier
	
	else:
		return '"{}"'.format(
			identifier.replace('"', '""'))


def compute_multi_join(sky_coords, radius):
	"""does an len(sky-coords)-way crossmatch of rows in sky_coords and returns
	the indices to match up.

	This is essentially an n-way symmetric crossmatch.  It's not efficient,
	though, and also exclusively positional.

	You probably want to use a real clustering algorithm here.  Astroml,
	http://www.astroml.org/, for instance, looks like a nice package.
	"""
	# match all tables against all tables for now
	matches = set()
	for left, right in itertools.permutations(range(len(sky_coords)), 2):
		idx, dist2, _ = sky_coords[left
			].match_to_catalog_sky(sky_coords[right])

		pairs = np.array([np.arange(len(idx)), idx]).transpose()
		for left_ind, right_ind in pairs[dist2<radius]:
			matches.add(((left, left_ind), (right, right_ind)))

	# aggregate the matches (i.e., put everything matched to the same thing
	# in one box
	to_join = {}
	for ob1, ob2 in matches:
		if (ob1 in to_join and ob2 in to_join 
				and to_join[ob1] is not to_join[ob2]):
			to_join[ob1] = to_join[ob2] = to_join[ob1] | to_join[ob2]
		elif ob2 in to_join:
			to_join[ob1] = to_join[ob2]
			to_join[ob1].add(ob1)
		elif ob1 in to_join:
			to_join[ob2] = to_join[ob1]
			to_join[ob2].add(ob2)
		else:
			to_join[ob1] = to_join[ob2] = set([ob1, ob2])
	
	return set(frozenset(v) for v in to_join.values())


def run_sync_resilient(svc, *sync_args, **sync_kw_args):
	"""runs a sync query in a TAP service svc, catching and logging all 
	kinds of errors.

	On error, this just returns None.

	In particular, this catches ^C so people can cancel individual
	queries.

	This is really intended for all-VO-queries where we don't and shouldn't
	care about a couple of broken services.
	"""
	try:
		return svc.run_sync(*sync_args, **sync_kw_args)  
	except (
			pyvo.dal.DALServiceError, 
			pyvo.dal.DALQueryError,
			requests.ConnectionError) as ex:
		print("{}:{}".format(svc.baseurl, ex))
		return
	except KeyboardInterrupt: # Let the user abort slow queries
		return
